# IR_RX_NEC.py - 20.06.2021

from machine import Pin, lightsleep
import utime as time
import uarray as arr

# Es werden 32 Datenbit übertragen
# Das zweit-höchstwertige Byte [2] wird als Datenbyte interpretiert,
# Das höchstwertige Byte [3] ist das invertierte Datenbyte.
# das niedrigwertigste Byte [0]als Adressbyte.
# das zweitniedrigste Byte [1] ist das invertierte Adressbyte.
# 
# 32 Bit Daten  Byte[3]     Byte[2]    Byte[1]    Byte[0]
#               inv. Daten  Daten      inv. Adr.  Adr

class IR_Nec(object):
    def __init__(self, pin, adr=None, cb_data=None, cb_repeat=None, debug=True):
        self.cb_data = cb_data
        self.cb_repeat = cb_repeat
        self.adr = adr
        self.repeated = 0
        self.debug = debug
        self.rx = 0
        self.cnt = 0
        self.ir_byte = arr.array('H', [0] * 4)
        self.last = time.ticks_us()
        self.ir_pin = Pin(pin, Pin.IN, Pin.PULL_UP)
        self.ir_pin.irq(self.read_pin, trigger=(Pin.IRQ_FALLING)) 

    def read_pin(self, _):
        now = time.ticks_us()
        pulse =  time.ticks_diff(now, self.last)
        self.last = now
        
        if pulse > 15000 and self.cnt:
            self._ir_reset()
            return
        
        if self.cnt == 0:
            # IRQ auf Rising Edge umstellen
            self.ir_pin.irq(self.read_pin, trigger=(Pin.IRQ_RISING)) 

        elif self.cnt == 1:
            # Startbit Pulse: 8000us
            if not (pulse > 8500 and pulse < 9500):
                self._ir_reset()
                if self.debug:
                    print("Error Start", pulse)
                return
            
        elif self.cnt == 2:
            # Startbit Pause + Pulse: 4500us + 560us
            if (pulse > 4000 and pulse < 5400):
                pass
            
            # Repeatbit Pause 2250us + 560us
            elif (pulse > 2500 and pulse < 3200):
                if self.cb_repeat:
                    self.repeated += 1
                    self.cb_repeat(self.repeated, self.ir_byte[2])
                    
                self._ir_reset()
                return 
 
            else:
                self._ir_reset()
                if self.debug:
                    print("Error ", self.cnt, pulse)
                return
            
        elif self.cnt < 35 :
            # Pause + Pulse (des nachfolgenden Bit) 560us* 2 = 1120
            if pulse > 1050 and pulse <1250:
                self.rx >>= 1# 
                
            elif pulse > 2100 and pulse < 2350: # 560 + 1690 = 2250
                self.rx >>= 1
                self.rx |= 0x80000000

            else: 
                self._ir_reset()
                if self.debug:
                    print("Error bit:", self.cnt, pulse)
                return
                    
            if self.cnt == 34:
                # Stopbit abgelaufen
                self._ir_analyse()
                return
                    
        self.cnt += 1
    
    def _ir_analyse(self):
        error = 0
        self.repeated = 0
        # Adresse und Command betehen aus komplementären Bytes
        # die Summe muss 255 * 2 betragen
        sum_bytes = 0
        for i in range(4):
            self.ir_byte[i] = self.rx & 0xFF
            sum_bytes += self.ir_byte[i]
            self.rx >>= 8   

        if not sum_bytes == 510:
            error |= 1
            if self.debug:
                print("SumErr: ", sum_bytes)
        
        # ggf. Adresse prüfen
        if self.adr is not None:
            if not self.adr == self.ir_byte[0]:
                error |= 2
                if self.debug:
                    print("Error: fixed Adr is: {} but received: {}".format(self.adr, self.ir_byte[0]))
                    
        self._ir_reset()
        if self.cb_data is not None:
            self.cb_data((self.ir_byte[0], self.ir_byte[2], error))
        else:
            print("Adr: {}|{} Cmd: {}|{}, Err: {}".format(self.ir_byte[0], self.ir_byte[1], self.ir_byte[2], self.ir_byte[3], error))
    
    def _ir_reset(self):
        self.cnt = 0
        self.rx = 0
        self.ir_pin.irq(self.read_pin, trigger=(Pin.IRQ_FALLING))

if __name__ == "__main__":

    # Callback zum Abholen der eingegangenen Daten
    def read_data(r):
        print("Message from callback: Adr: {}, Cmd: {}, Err: {}".format(r[0], r[1], r[2]))
        # tu was mit den Daten
        
    # Callback zum Registrieren von längeren Tastendrücken
    def read_repeated(x, cmd):
        print("- Cmd: {} repeated: {}".format(cmd, x))
        # tu was mit den Daten
      
    ir_pin = 17
    adr   = 128
    
    print("IR-Receiver mit NEC-Protokoll an Pin {}, Adresse: {}\n".format(ir_pin, adr))
    ir = IR_Nec(pin=ir_pin, adr=None, cb_data=read_data, cb_repeat=read_repeated, debug = True) 

    while True:
        lightsleep()
