# IR_RX_NEC_PIO.py  -  22.06.2021

import time
import rp2
from machine import Pin

# Es werden 32 Datenbit übertragen
# Das zweit-höchstwertige Byte [2] wird als Datenbyte interpretiert,
# Das höchstwertige Byte [3] ist das invertierte Datenbyte.
# das niedrigwertigste Byte [0] als Adressbyte.
# das zweitniedrigste Byte [1] ist das invertierte Adressbyte,
# 
# 32 Bit Daten  Byte[3]     Byte[2]    Byte[1]    Byte[0]
#               inv. Daten  Daten      inv. Adr.  Adr
# -----------------------------------------

@rp2.asm_pio(set_init=rp2.PIO.OUT_LOW)
# Hier wird der 38KHz Takt für die IR-Led generiert
def led():
    irq(4)                   # IRQ 4 setzen, blocking
    
    wrap_target()
    set(pins, 0)       [3]
                             # warten, falls IRQ 4 gesetzt ist
    wait (0, irq, 4)         # 0 = das Flag wird nicht gelöscht
    set (pins, 1)      [4] 
    wrap()
    
# ---------------------------------------

@rp2.asm_pio(out_shiftdir=rp2.PIO.SHIFT_RIGHT, fifo_join=rp2.PIO.JOIN_TX)

# Zum Debuggen des Timings mit einem zusätzlichem Pin, diesen Pin (set(pin, x) im Code freigeben 
# und ein nop() am Ende auskommentieren (sonst reicht der Speicher nicht aus).
# @rp2.asm_pio(out_shiftdir=rp2.PIO.SHIFT_RIGHT, set_init=rp2.PIO.OUT_HIGH, fifo_join=rp2.PIO.JOIN_TX)

# hier werden das Startbit, die Datenbits und das Stopbit moduliert,
# indem per IRQ 4 die IR-Diode ein- und ausgeschaltet wird.
# TX_FIFO ist um RX_FIFO erweitert (auf 8 Worte)
# Träger  : 38 KHz
# Takt    : 20us / 50_000 Hz
# Startbit 8000 us (160 Takte)  + 4500 us (90 Takte)
# Bit_0    = 560 + 560 us   28 * 20 us =  560 us
# Bit_1    = 560 + 1690 us  85 * 20 us = 1700 us
# Stop     = 560 us + 30ms Wartezeit
# [Wiederholung: 8000us + 2250 us + 560 us + 100 ms Pause]

def ir_nec():
    wrap_target() 
    pull()                     # blocking, warten auf FIFO
                               # das Datenwort (32Bit) ist im OSR                      
    set(y, 7)                  # 8 * 1000 us

    #set(pins, 0)  # ---------- (stört das timing nicht, verzögert um 20 us)
    irq(clear, 4)              # Träger ein für Startbit
 
    label("delay_1")           # 8 * 1000 us
    nop()                 [24] 
    jmp(y_dec, "delay_1") [24] # 50 Takte je 20 us * 8 = 800 us
 
    #set(pins, 1)  # ---------- (verlängert das Startbit um 20 us) 
    irq(4)                     # Träger aus Startbit
    
    set(y, 8)                  # 9 * 500 us
    label("delay_2")           
    jmp(y_dec, "delay_2") [24] # 25 Takte je 20 us ? 8 = 450us
    
    label("next_bit")
 
    #set(pins, 0)  # ---------- (verlängert die Pausezeiten um 20 us)
    irq(clear, 4)              # Ende der PauseTräger ein
    
    out(x, 1)             [26] # das nächste Bit nach x
                               # bis hier 28 Takte je 20 us = 560 us
    #set(pins, 1)  # ---------- (verlängert Pulsezeit um 20 us)
    irq(4)                [25] # Träger aus für 26 (+1 +1) * 20 us = 560 us

    
    jmp(not_x,"no_nops")       # wenn 0, dann die mächsten nops auslassen
    
    nop()                 [25] # 1690 - 560 = 1130
    nop()                 [29] # 56 (+ 1) * 20 =  1140
    
    label("no_nops")
    jmp(not_osre, "next_bit")  # weiter solange wie Bits vorhanden sind

    #set(pins, 0)  # ---------- (verlängert die Pausezeiten um weitere 20 us
    irq(clear, 4)              # Träger ein für 560 us

    nop()                 [26] # 27 (+1) * 20  = 560 us
    #set(pins, 1)  # ---------- (verlängert Pulse um 20 us
    irq(4)                     # Träger aus

    
    set(y, 31)                 # Pause nach einer übertragung einlegen
    label("delay_3")           # ca. 140ms Abstand zur ggf. nächsten Sognal
                               # willkürlich gewählt !
    nop()                 [31]
    nop()                 [31]
    nop()                 [31] # ((32 * 4) * 20 us = 4.4ms) * 32 = 140ms
    jmp(y_dec, "delay_3") [31] 
                               # nur als Test ohne weiteren Sinn
    #irq(rel(0))               # löst einen Interrupt am Ende der Übertragung aus
                               # führt den irq_handler aus - falls gewünscht
    wrap()  
   
# --------------------------------------------------------

class IR_TX(object):
    def __init__(self, pin, irq_handler=None, freq=38_000):
        self.sm0 = rp2.StateMachine(0, led, freq=freq*10, set_base=Pin(pin))
        # 380.000 / 10 Takte = 38.000 Hz 
  
        self.sm1 = rp2.StateMachine(1, ir_nec, freq=50_000)
        # Takt 20 us
        # zum Debugen des Timings zusätzlichen Pin verwenden
        # self.sm1 = rp2.StateMachine(1, ir_nec, freq=50_000, set_base=Pin(2))

        
        # Einen Interrupt ausführen nach einer Übertragung - Nur zum Testen
        if irq_handler is not None:
            self.sm1.irq(irq_handler)

        self.sm0.active(1)
        self.sm1.active(1)
 
    def send(self, data):
        if not type(data) is bytes:
            print("Error: expecting bytes")
            return
        elif not len(data) == 2:
            print("Error: expecting 2 bytes")
            return
        
        msg = (255 - data[1]) * 256 + data[1]                # cmd
        msg = msg * 256**2 + (255 - data[0]) * 256 + data[0] # adr

        # print("{:08x}".format(msg))
        self.sm1.put(msg)
    
    def words_in_fifo(self):
        return self.sm1.tx_fifo()
      

if __name__ == "__main__":
    # Output-Pin für die IR-LED
    ir_tx_pin = 0
 
    adr = 128
    cmd = 13
    
    # nur als Test
    def irq_handler(_):
        x = tx.words_in_fifo()
        print("Frame out, Fifo free:", x)
        
    tx = IR_TX(pin=ir_tx_pin, irq_handler=irq_handler)
    tx.send(bytes([adr, cmd]))
   
    time.sleep(2)

    print("done")
